version: '3.7'
services:
  portainer-agent:
    image: {{ ipv6_registry }}portainer/agent
    container_name: portainer-agent
    restart: unless-stopped
    volumes:
      - /var/run/docker.sock:/var/run/docker.sock
      - /var/lib/docker/volumes:/var/lib/docker/volumes
    expose:
      - 9001
    ports:
      - "9001:9001"
    network_mode: bridge

  dtnaas-agent:
    image: {{ ipv6_registry }}dtnaas/agent
    container_name: dtnaas-agent
    restart: unless-stopped
    privileged: true
    environment:
      - AGENT_BIND=0.0.0.0
      - AGENT_PORT=5050
      - AGENT_SSL=--ssl
    expose:
      - 5050
    ports:
      - "5050:5050"
    network_mode: host
