CLOUDLAB_PROJECT = "project"
CLOUDLAB_CERTIFICATE = "certificate"
CLOUDLAB_USER = "user"
CLOUDLAB_SLICE_PRIVATE_KEY_LOCATION = "slice-private-key-location"

CLOUDLAB_CONF_ATTRS = [CLOUDLAB_PROJECT, CLOUDLAB_CERTIFICATE, CLOUDLAB_USER, CLOUDLAB_SLICE_PRIVATE_KEY_LOCATION]
CLOUDLAB_RETRY = 100

CLOUDLAB_SLEEP_TIME = 5
CLOUDLAB_MAX_EXPERIMENT_NAME_SIZE = 16
AGGREGATE_STATUS = "aggregate_status"
NODES = 'nodes'
NODE = 'node'

# {'ipv4': '155.98.36.152', 'hostname': 'node.aes-cloudlab.amscrot.emulab.net'}
IPV4 = "ipv4"
HOSTNAME = "hostname"
