ACCESS_KEY = "ACCESS_KEY"
SECRET_KEY = "SECRET_KEY"
RETRY = 60

VLAN = 'vlan'
AMAZON_SIDE_ASN = 'amazonSideAsn'
ASN = 'asn'
VIF_ID = 'virtualInterfaceId'
VIF_STATE = 'virtualInterfaceState'

VIF_DETAILS = [VLAN, AMAZON_SIDE_ASN, ASN, VIF_ID, VIF_STATE]
DO_NOT_USE_MAIN_ROUTING_TABLE = True
